#!/usr/bin/env node
"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const readChunk = __importStar(require("read-chunk"));
const https_1 = __importDefault(require("https"));
const inquirer_autocomplete_prompt_1 = __importDefault(require("inquirer-autocomplete-prompt"));
const inquirer = require('inquirer');
const fs = __importStar(require("fs"));
const path_1 = __importDefault(require("path"));
const chalk = __importStar(require("chalk"));
const toml = __importStar(require("@iarna/toml"));
const os = __importStar(require("os"));
const cross_spawn_1 = __importDefault(require("cross-spawn"));
const lodash_1 = require("lodash");
const ora_1 = __importDefault(require("ora"));
const temp = __importStar(require("temp"));
const moment_1 = __importDefault(require("moment"));
const inquirer_directory_1 = __importDefault(require("inquirer-directory"));
const fuzzy_1 = require("fuzzy");
const child_process_1 = require("child_process");
const appdata_path_1 = __importDefault(require("appdata-path"));
const request_1 = __importDefault(require("request"));
const cli_table3_1 = __importDefault(require("cli-table3"));
const clikey_1 = __importDefault(require("clikey"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const command_line_args_1 = __importDefault(require("command-line-args"));
const command_line_usage_1 = __importDefault(require("command-line-usage"));
const term_size_1 = __importDefault(require("term-size"));
const tiny_glob_1 = __importDefault(require("tiny-glob"));
const diff_1 = __importDefault(require("diff"));
class DiffPatch {
    constructor() {
        this.oDefaultConfigInfo = {
            DiffMethod: 'normal',
            Project: [
                {
                    Name: 'Code Compare',
                    LeftPath: '/home/code1',
                    RightPath: '/home/jason/code2',
                    FilePattern: '**/*!({pdf,doc,node_modules})*',
                    ExcludeFilePattern: '',
                    ExcludeFolderPattern: '',
                    NoQuestions: 'false',
                },
                {
                    Name: 'Text Compare',
                    LeftPath: '/home/text1',
                    RightPath: '/home/jason/text2',
                    FilePattern: '',
                    ExcludeFilePattern: '*.jpg;*.png;*.pdf',
                    ExcludeFolderPattern: 'node_modules',
                    NoQuestions: 'false',
                },
            ],
        };
        this.strTomlConfigFolder = os.homedir() + '/.interactive-diff-patch/';
        this.strTomlConfigFile = this.strTomlConfigFolder + 'config.toml';
        this.strLeftPath = '';
        this.strRightPath = '';
        this.strFileExcludes = '';
        this.strFolderExcludes = '';
        this.strArrListOfFileChanges = new Array();
        this.strNanoGitWinPath = 'c:\\Program Files\\Git\\usr\\bin\\nano.exe';
        this.strDiffGitWinPath = 'c:\\Program Files\\Git\\usr\\bin\\diff.exe';
        this.strPatchGitWinPath = 'c:\\Program Files\\Git\\usr\\bin\\patch.exe';
        this.terminalSize = term_size_1.default();
        const optionDefinitions = [
            { name: 'LeftFolder', alias: 'l', type: String, description: 'Left Folder - the one with the changes.' },
            { name: 'RightFolder', alias: 'r', type: String, description: 'Right Folder.' },
            {
                name: 'XFiles',
                alias: 'x',
                type: String,
                description: 'File patters to ignore.  Example: *.txt;*.png;*.bak',
            },
            {
                name: 'XFolders',
                alias: 'f',
                type: String,
                description: 'Folder patters to ignore.  Example: .git;node_modules;test*',
            },
            {
                name: 'NoQuestions',
                alias: 'q',
                type: String,
                description: "Set to 'y' to skip questions.  Default is 'n'.",
            },
            { name: 'help', alias: 'h', description: 'View help.' },
        ];
        const sections = [
            {
                header: 'Bixfliz idiff-patch (interactive diff and merge/patch)',
                content: 'Bixfliz interactive folder compare and merge tool.  Start with no command line options to run in interactive mode.  Projects and settings can be set in the config located file at ' +
                    this.strTomlConfigFile +
                    '.  The DiffMethod option in the config file can be "normal", "diff", or "robocopy" (robocopy is only for Windows).  \n\nExample:\nidiff-patch -l /var/test-code -r /var/code -f node_modules;.git -q y',
            },
            {
                header: 'Options',
                optionList: optionDefinitions,
            },
        ];
        this.options = command_line_args_1.default(optionDefinitions);
        if (this.options.help === null) {
            console.log(command_line_usage_1.default(sections));
            process.exit();
        }
        if (!fs.existsSync(this.strTomlConfigFolder + '/')) {
            fs.mkdirSync(this.strTomlConfigFolder + '/');
        }
        if (!fs.existsSync(this.strTomlConfigFile)) {
            fs.closeSync(fs.openSync(this.strTomlConfigFile, 'w'));
            fs.writeFileSync(this.strTomlConfigFile, toml.stringify({
                Project: this.oDefaultConfigInfo.Project,
            }), { flag: 'w+' });
        }
        if (os.type() === 'Windows_NT') {
            this.getNano().then(() => {
                this.selectCompareFolders();
            });
        }
        else {
            this.selectCompareFolders();
        }
    }
    uuidv4() {
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
            var r = (Math.random() * 16) | 0, v = c == 'x' ? r : (r & 0x3) | 0x8;
            return v.toString(16);
        });
    }
    backupFile(Fullpath) {
        let strAppDataFolder = appdata_path_1.default('interactive-diff-patch');
        let strAppDataBackupFolder = path_1.default.join(strAppDataFolder, 'backups');
        if (!fs.existsSync(strAppDataFolder)) {
            fs.mkdirSync(strAppDataFolder);
        }
        if (!fs.existsSync(strAppDataBackupFolder)) {
            fs.mkdirSync(strAppDataBackupFolder);
        }
        let strFilename = path_1.default.basename(Fullpath);
        let strBackupPath = path_1.default.join(strAppDataBackupFolder, this.uuidv4() + strFilename);
        fs.copyFileSync(Fullpath, strBackupPath);
        console.log('File backed up to: ' + strBackupPath);
    }
    async getNano() {
        return new Promise((resolve, reject) => {
            setTimeout(() => {
                try {
                    if (fs.existsSync(this.strNanoGitWinPath)) {
                        resolve();
                    }
                    let strAppDataFolder = appdata_path_1.default('interactive-diff-patch');
                    let strDownloadFilePath = strAppDataFolder + '/nano.exe';
                    this.strTomlConfigFile = this.strTomlConfigFile.replace(/\\/g, '/');
                    let booDownloadExists = fs.existsSync(strDownloadFilePath);
                    if (booDownloadExists) {
                        resolve();
                    }
                    else {
                        if (!fs.existsSync(strAppDataFolder)) {
                            fs.mkdirSync(strAppDataFolder);
                        }
                        let strBaseUrl = 'https://www.nano-editor.org/dist/win32-support/';
                        this.getWebPage(strBaseUrl).then((strHtml) => {
                            let oMatch = /<a href="(nano-git-[0-9A-z]+\.exe)">/gi.exec(strHtml);
                            let strNanoWinUrl = '';
                            if (!fs.existsSync(strAppDataFolder)) {
                                fs.mkdirSync(strAppDataFolder);
                            }
                            if (oMatch && oMatch[1]) {
                                strNanoWinUrl = oMatch[1];
                            }
                            if (strNanoWinUrl) {
                                this.download(strBaseUrl + strNanoWinUrl, strDownloadFilePath, (msg) => {
                                    if (msg) {
                                        console.log(msg);
                                        reject(msg);
                                    }
                                    else {
                                        resolve();
                                    }
                                });
                            }
                        });
                    }
                }
                catch (error) {
                    reject(error);
                }
            });
        });
    }
    async download(url, dest, cb) {
        var file = fs.createWriteStream(dest);
        var sendReq = request_1.default.get(url);
        // verify response code
        sendReq.on('response', function (response) {
            if (response.statusCode !== 200) {
                return cb('Response status was ' + response.statusCode);
            }
        });
        // check for request errors
        sendReq.on('error', function (err) {
            fs.unlinkSync(dest);
            return cb(err.message);
        });
        sendReq.pipe(file);
        file.on('finish', function () {
            file.close(); // close() is async, call cb after close completes.
            return cb('');
        });
        file.on('error', function (err) {
            // Handle errors
            fs.unlinkSync(dest); // Delete the file async. (But we don't check the result)
            return cb(err.message);
        });
    }
    async selectCompareFolders() {
        this.strLeftPath = '';
        this.strRightPath = '';
        this.strFileExcludes = '';
        this.strFolderExcludes = '';
        let strData = fs.readFileSync(this.strTomlConfigFile).toString();
        let oData = toml.parse(strData);
        this.oConfigInfo = oData;
        let oCliTable = new cli_table3_1.default({ style: { head: [], border: [] } });
        if (this.options.FolderA && this.options.FolderB) {
            this.strLeftPath = this.options.FolderA;
            this.strRightPath = this.options.FolderB;
            let strFileExcludes = '';
            let strFolderExcludes = '';
            let strNoQuestions = '';
            if (this.options.XFiles)
                strFileExcludes = this.options.XFiles;
            if (this.options.XFolders)
                strFolderExcludes = this.options.XFolders;
            if (this.options.NoQuestions && this.options.NoQuestions.toLowerCase() == 'y') {
                strNoQuestions = 'true';
            }
            else {
                strNoQuestions = 'false';
            }
            this.oConfigProject = {
                Name: 'Args',
                LeftPath: this.strLeftPath,
                RightPath: this.strRightPath,
                FilePattern: '',
                ExcludeFilePattern: strFileExcludes,
                ExcludeFolderPattern: strFolderExcludes,
                NoQuestions: strNoQuestions,
            };
        }
        else {
            oCliTable.push([{ colSpan: 2, content: chalk.default.cyanBright('Press a Key to Select Action:') }], [chalk.default.greenBright('p'), 'Start by selecting a project'], [chalk.default.greenBright('m'), 'Enter Values Manually'], [chalk.default.greenBright('e'), 'Edit system config file']);
            console.log('\n\n');
            console.log(oCliTable.toString());
            let strLetter = await clikey_1.default('?');
            console.log('>' + strLetter);
            console.log('\n\n');
            if (strLetter.startsWith('e')) {
                this.editFileSync(this.strTomlConfigFile);
                this.selectCompareFolders();
                return;
            }
            else if (strLetter.startsWith('p')) {
                let strArrProject = new Array();
                for (let oProject of this.oConfigInfo.Project) {
                    strArrProject.push(oProject.Name);
                }
                console.log('\n\n');
                let strProject = await this.askForListChoiceFuzzy('Project:', strArrProject);
                for (var intEachItem = 0; intEachItem < this.oConfigInfo.Project.length; intEachItem++) {
                    let oProject = this.oConfigInfo.Project[intEachItem];
                    if (oProject.Name == strProject) {
                        this.oConfigProject = oProject;
                        this.strFileExcludes = this.oConfigProject.ExcludeFilePattern;
                        this.strFolderExcludes = this.oConfigProject.ExcludeFolderPattern;
                        this.strLeftPath = this.oConfigProject.LeftPath;
                        this.strRightPath = this.oConfigProject.RightPath;
                        if (!this.oConfigProject ||
                            (this.oConfigProject.NoQuestions && this.oConfigProject.NoQuestions != 'true')) {
                            console.log('\n\n');
                            this.strFileExcludes = await this.askForString('Exclude file pattern: ', this.strFileExcludes);
                            console.log('\n\n');
                            this.strFolderExcludes = await this.askForString('Exclude folder pattern: ', this.strFolderExcludes);
                        }
                        // this.selectFile()
                        break;
                    }
                }
            }
            else if (strLetter.startsWith('m')) {
                console.log('\n\n');
                this.strFileExcludes = await this.askForString('Exclude file patteren (e.g.: *.png;*.jpg;*.pdf;): ', '');
                console.log('\n\n');
                this.strFolderExcludes = await this.askForString('Exclude file patteren (e.g.: node_modules;*-test): ', '');
            }
        }
        if (this.strLeftPath) {
            if (!fs.existsSync(this.strLeftPath)) {
                console.error(this.strLeftPath + ' not found.');
                this.strLeftPath = '/';
            }
        }
        else {
            this.strLeftPath = '/';
        }
        if (this.strRightPath) {
            if (!fs.existsSync(this.strRightPath)) {
                console.error(this.strRightPath + ' not found.');
                this.strRightPath = '/';
            }
        }
        else {
            this.strRightPath = '/';
        }
        let strFolderSep = '/';
        if (os.type() === 'Windows_NT') {
            strFolderSep = '\\';
            this.strLeftPath = this.strLeftPath.replace(/\//g, '\\');
            this.strLeftPath = this.strLeftPath.replace(/\\\\/g, '\\');
            this.strRightPath = this.strRightPath.replace(/\//g, '\\');
            this.strRightPath = this.strRightPath.replace(/\\\\/g, '\\');
        }
        else {
            this.strLeftPath = this.strLeftPath.replace(/\\/g, '/');
            this.strLeftPath = this.strLeftPath.replace(/\/\//g, '/');
            this.strRightPath = this.strRightPath.replace(/\\/g, '/');
            this.strRightPath = this.strRightPath.replace(/\/\//g, '/');
        }
        if (!this.strLeftPath.endsWith(strFolderSep))
            this.strLeftPath = this.strLeftPath + strFolderSep;
        if (!this.strRightPath.endsWith(strFolderSep))
            this.strRightPath = this.strRightPath + strFolderSep;
        if ((this.oConfigProject && this.oConfigProject.NoQuestions && this.oConfigProject.NoQuestions != 'true') ||
            !this.oConfigProject) {
            if (os.type() === 'Windows_NT') {
                if (this.strLeftPath == '/') {
                    this.strLeftPath = os.homedir(); // + '\\'
                }
                else if (this.strLeftPath == '\\') {
                    this.strLeftPath = os.homedir(); // + '\\'
                }
                if (this.strRightPath == '/') {
                    this.strRightPath = os.homedir(); // + '\\'
                }
                else if (this.strRightPath == '\\') {
                    this.strRightPath = os.homedir(); // K+ '\\'
                }
            }
            let strSelectedSubPathLeft;
            let strSelectedSubPathRight;
            console.log('\n\n');
            strSelectedSubPathLeft = await this.askForFolder('Select first folder:', this.strLeftPath);
            console.log('\n\n');
            strSelectedSubPathRight = await this.askForFolder('Select second folder:', this.strRightPath);
            this.strLeftPath = path_1.default.join(this.strLeftPath, strSelectedSubPathLeft);
            this.strRightPath = path_1.default.join(this.strRightPath, strSelectedSubPathRight);
            this.loadFolderDifferences();
        }
        else {
            this.loadFolderDifferences();
        }
    }
    async loadFolderDifferences() {
        if (this.oConfigInfo.DiffMethod == 'robocopy') {
            await this.loadFolderDifferencesWithDiff(true);
            return;
        }
        else if (this.oConfigInfo.DiffMethod == 'diff') {
            await this.loadFolderDifferencesWithDiff();
            return;
        }
        const spinner = ora_1.default({
            text: 'Loading differences...',
            color: 'yellow',
            spinner: 'circle',
        });
        spinner.start();
        let strExcludes = '';
        if (this.oConfigProject.FilePattern) {
            strExcludes = this.oConfigProject.FilePattern;
        }
        else {
            strExcludes = "**/*";
        }
        let strFolderPath = this.strLeftPath.replace(/\\/g, '/');
        // let strFiles: string[] = await glob('**/*!({pdf,js,png,DoNotPublish,.vs,bak,gif,Obsolete,cnt,dll,doc.docx,dot,exe,frx,gif,Thumbs.db,mdb,jpg,png,ppt,config,bak,log,fpdb,ADEQWebsiteEditor,Intranet,obj,bin,csv,node_modules,.git,ctx,mp3,mp4,mov,wav})*', { cwd: strFolderPath, filesOnly: true });
        setTimeout(() => {
            tiny_glob_1.default(strExcludes, {
                cwd: strFolderPath,
                filesOnly: true,
            }).then((strFiles) => {
                let oTimerStart = moment_1.default();
                this.strArrListOfFileChanges = new Array();
                for (let i = 0; i < strFiles.length; i++) {
                    let strFileRelativePath = strFiles[i];
                    let fullPath = path_1.default.join(this.strLeftPath, strFileRelativePath);
                    let strFileTwoFullPath = this.strRightPath + strFileRelativePath;
                    let stats1 = fs.statSync(fullPath); //, (err, stats1) => {
                    if (fs.existsSync(strFileTwoFullPath)) {
                        let stats2 = fs.statSync(strFileTwoFullPath); //, (err, stats2) => {
                        let nTimeDiffInSeconds = Math.abs((stats1.mtime.getTime() - stats2.mtime.getTime()) / 1000);
                        if (stats1.size > stats2.size) {
                            this.strArrListOfFileChanges.push('   Bigger: ' + strFileRelativePath);
                        }
                        else if (stats1.size < stats2.size) {
                            this.strArrListOfFileChanges.push('  Smaller: ' + strFileRelativePath);
                        }
                        else if (nTimeDiffInSeconds > 1 && stats1.mtime > stats2.mtime) {
                            let i = 0;
                            let oData1;
                            let oData2;
                            do {
                                oData1 = readChunk.sync(fullPath, i * 10000, 10000);
                                oData2 = readChunk.sync(strFileTwoFullPath, i * 10000, 10000);
                                if (oData1.compare(oData2) != 0) {
                                    this.strArrListOfFileChanges.push('    Newer: ' + strFileRelativePath);
                                    break;
                                }
                                i++;
                            } while (oData1.byteLength && oData2.byteLength && oData1.byteLength == 10000 && oData2.byteLength == 10000);
                        }
                        else if (nTimeDiffInSeconds > 1 && stats1.mtime < stats2.mtime) {
                            let i = 0;
                            let oData1;
                            let oData2;
                            do {
                                oData1 = readChunk.sync(fullPath, i * 10000, 10000);
                                oData2 = readChunk.sync(strFileTwoFullPath, i * 10000, 10000);
                                if (oData1.compare(oData2) != 0) {
                                    this.strArrListOfFileChanges.push('    Older: ' + strFileRelativePath);
                                    break;
                                }
                                i++;
                            } while (oData1.byteLength && oData2.byteLength && oData1.byteLength == 10000 && oData2.byteLength == 10000);
                        }
                    }
                    else {
                        this.strArrListOfFileChanges.push('< <  Only: ' + strFileRelativePath);
                    }
                }
                if (this.strArrListOfFileChanges &&
                    this.strArrListOfFileChanges.length &&
                    this.strArrListOfFileChanges[this.strArrListOfFileChanges.length - 1].trim().length == 0)
                    this.strArrListOfFileChanges.pop();
                let oTimerEnd = moment_1.default();
                let intMilliseconds = oTimerEnd.diff(oTimerStart, 'milliseconds');
                let intSeconds = Math.round(intMilliseconds / 100) / 10;
                let intMinutes = Math.round(intMilliseconds / 100 / 60) / 10;
                let strFileCount = this.strArrListOfFileChanges.length.toLocaleString();
                let strStatusLine = '';
                if (intSeconds == 1) {
                    strStatusLine = strFileCount + ' files in 1 second.';
                }
                else if (intSeconds < 60) {
                    strStatusLine = strFileCount + ' files in ' + intSeconds + ' seconds.';
                }
                else if (intSeconds == 60) {
                    strStatusLine = strFileCount + ' files in 1 minute.';
                }
                else {
                    strStatusLine = strFileCount + ' files in ' + intMinutes.toLocaleString() + ' minutes.';
                }
                spinner.text = strStatusLine;
                spinner.succeed();
                this.showFolderDifferences();
            });
        }, 1000);
    }
    async loadFolderDifferencesWithDiff(UseRobocopy) {
        let strCommand = '';
        let strArrFolderDiffSwitches = ['--brief', '--recursive'];
        let strFolderDiffTempFilename = temp.path({ suffix: '.diff' });
        let strDiffProgramExcludeFilename = temp.path({ suffix: '-excludes.txt' });
        if (UseRobocopy) {
            //
            strCommand = 'robocopy';
            strArrFolderDiffSwitches = [
                this.strLeftPath,
                this.strRightPath,
                '/log:' + strFolderDiffTempFilename,
                '/L',
                '/e',
                '/ns',
                '/njs',
                '/njh',
                '/ndl',
                '/fp',
                '/np',
                '/fft',
            ];
            if (this.strFileExcludes) {
                this.strFileExcludes = this.strFileExcludes.replace(/\;/g, ' ');
                this.strFileExcludes = this.strFileExcludes.replace(/\[\)\]/g, ')');
                this.strFileExcludes = this.strFileExcludes.replace(/\[\(\]/g, '(');
                strArrFolderDiffSwitches.push('/XF');
                strArrFolderDiffSwitches = strArrFolderDiffSwitches.concat(this.strFileExcludes.split(' '));
            }
            if (this.strFolderExcludes) {
                this.strFolderExcludes = this.strFolderExcludes.replace(/\;/g, ' ');
                this.strFolderExcludes = this.strFolderExcludes.replace(/\[\)\]/g, ')');
                this.strFolderExcludes = this.strFolderExcludes.replace(/\[\(\]/g, '(');
                strArrFolderDiffSwitches.push('/XD');
                strArrFolderDiffSwitches = strArrFolderDiffSwitches.concat(this.strFolderExcludes.split(' '));
            }
        }
        else {
            strCommand = 'diff';
            let strExcludeListForDiffProg = '';
            if (this.strFileExcludes && this.strFolderExcludes) {
                strExcludeListForDiffProg = this.strFileExcludes + ';' + this.strFolderExcludes;
                strExcludeListForDiffProg = strExcludeListForDiffProg.replace(/\;/g, '\n');
                strExcludeListForDiffProg = strExcludeListForDiffProg.replace(/\n\n/g, '\n');
            }
            fs.writeFileSync(strDiffProgramExcludeFilename, strExcludeListForDiffProg, { flag: 'w+' });
            strArrFolderDiffSwitches.push('-X ' + strDiffProgramExcludeFilename);
            strArrFolderDiffSwitches.push(this.strLeftPath);
            strArrFolderDiffSwitches.push(this.strRightPath);
            strArrFolderDiffSwitches.push('>');
            strArrFolderDiffSwitches.push(strFolderDiffTempFilename);
        }
        console.log(chalk.default.grey.dim.italic('\n\nRunning: ' + strCommand + ' ' + strArrFolderDiffSwitches.join(' ')));
        console.log('\n\n');
        const spinner = ora_1.default({
            text: 'Loading differences...',
            color: 'yellow',
            spinner: 'circle',
        });
        spinner.start();
        let oTimerStart = moment_1.default();
        if (UseRobocopy) {
            //
            const oChildProcess = cross_spawn_1.default(strCommand, strArrFolderDiffSwitches);
            oChildProcess.stdout.on('error', (chunk) => {
                console.log(chunk.toString());
            });
            oChildProcess.stdout.on('data', (chunk) => {
                // console.log(chunk.toString())
            });
            oChildProcess.on('exit', async () => {
                let strFolderDiffOutput = fs.readFileSync(strFolderDiffTempFilename).toString();
                let strLeftPathForRegex = this.strLeftPath.replace(/\\/g, '\\\\');
                let strRightPathForRegex = this.strRightPath.replace(/\\/g, '\\\\');
                let oNewReplaceRegex = new RegExp('^New File +' + strLeftPathForRegex + '(.*?)$', 'gm');
                let oExtraReplaceRegex = new RegExp('^\\*EXTRA File +' + strRightPathForRegex + '(.*?)$', 'gm');
                let oOlderReplaceRegex = new RegExp('^Older +' + strLeftPathForRegex + '(.*?)$', 'gm');
                let oNewerReplaceRegex = new RegExp('^Newer +' + strLeftPathForRegex + '(.*?)$', 'gm');
                let oChangedReplaceRegex = new RegExp('^Changed +' + strLeftPathForRegex + '(.*?)$', 'gm');
                strFolderDiffOutput = strFolderDiffOutput.replace(/\t/g, '');
                strFolderDiffOutput = strFolderDiffOutput.replace(/\r/g, '');
                strFolderDiffOutput = strFolderDiffOutput.replace(/^ */gm, '');
                strFolderDiffOutput = strFolderDiffOutput.replace(/^$\n/gm, '');
                // strFolderDiffOutput = strFolderDiffOutput.replace(oOlderReplaceRegex, 'Different: $1')
                strFolderDiffOutput = strFolderDiffOutput.replace(oOlderReplaceRegex, '< < Older: $1');
                // strFolderDiffOutput = strFolderDiffOutput.replace(oNewerReplaceRegex, 'Different: $1')
                strFolderDiffOutput = strFolderDiffOutput.replace(oNewerReplaceRegex, ' > >Newer: $1');
                strFolderDiffOutput = strFolderDiffOutput.replace(oChangedReplaceRegex, 'Different: $1');
                strFolderDiffOutput = strFolderDiffOutput.replace(oNewReplaceRegex, '< <  Only: $1');
                strFolderDiffOutput = strFolderDiffOutput.replace(oExtraReplaceRegex, ' > > Only: $1');
                strFolderDiffOutput = strFolderDiffOutput.replace(/< <  Only: \//gm, '< <  Only: ');
                strFolderDiffOutput = strFolderDiffOutput.replace(/ > > Only: \//gm, ' > > Only: ');
                if (strFolderDiffOutput.trim()) {
                    let strArrChoices = strFolderDiffOutput.split('\n');
                    if (strArrChoices && strArrChoices[0].trim().length == 0)
                        strArrChoices.shift();
                    if (strArrChoices && strArrChoices[strArrChoices.length - 1].trim().length == 0)
                        strArrChoices.pop();
                    let oTimerEnd = moment_1.default();
                    let intMilliseconds = oTimerEnd.diff(oTimerStart, 'milliseconds');
                    let intSeconds = Math.round(intMilliseconds / 100) / 10;
                    let intMinutes = Math.round(intMilliseconds / 100 / 60) / 10;
                    let strFileCount = strArrChoices.length.toLocaleString();
                    let strStatusLine = '';
                    if (intSeconds == 1) {
                        strStatusLine = strFileCount + ' files in 1 second.';
                    }
                    else if (intSeconds < 60) {
                        strStatusLine = strFileCount + ' files in ' + intSeconds + ' seconds.';
                    }
                    else if (intSeconds == 60) {
                        strStatusLine = strFileCount + ' files in 1 minute.';
                    }
                    else {
                        strStatusLine = strFileCount + ' files in ' + intMinutes.toLocaleString() + ' minutes.';
                    }
                    spinner.text = strStatusLine;
                    spinner.succeed();
                    this.strArrListOfFileChanges = strArrChoices;
                }
                else {
                    spinner.text = 'No Differences';
                    spinner.succeed();
                }
                fs.unlink(strFolderDiffTempFilename, (err) => {
                    if (err)
                        console.log(err.message);
                });
                this.showFolderDifferences();
            });
        }
        else {
            var child = child_process_1.exec(strCommand + ' ' + strArrFolderDiffSwitches.join(' '));
            child.stdout.on('data', function (data) {
                // console.log('stdout: ' + data)
            });
            child.stderr.on('error', function (data) {
                console.log('eror: ' + data);
            });
            child.on('close', (code) => {
                let strFolderDiffOutput = fs.readFileSync(strFolderDiffTempFilename).toString();
                if (os.type() == 'Windows_NT') {
                    strFolderDiffOutput = strFolderDiffOutput.replace(/\//g, '\\');
                }
                let strLeftPathForRegex = this.strLeftPath.replace(/\\/g, '\\\\');
                let strRightPathForRegex = this.strRightPath.replace(/\\/g, '\\\\');
                // Different:
                // < <  Only:
                //  > > Only:
                let oDifferReplaceRegex = new RegExp('^Files ' + strLeftPathForRegex + '(.+) and ' + strRightPathForRegex + '.+ differ$', 'gm');
                let oLeftOnlyReplaceRegex = new RegExp('^Only in ' + strLeftPathForRegex + '(.*?): (.+?)$', 'gm');
                let oRightOnlyReplaceRegex = new RegExp('^Only in ' + strRightPathForRegex + '(.*?): (.+?)$', 'gm');
                strFolderDiffOutput = strFolderDiffOutput.replace(/\t/g, '');
                strFolderDiffOutput = strFolderDiffOutput.replace(/\r/g, '');
                strFolderDiffOutput = strFolderDiffOutput.replace(/^$\n/gm, '');
                strFolderDiffOutput = strFolderDiffOutput.replace(oDifferReplaceRegex, 'Different: $1');
                if (os.type() == 'Windows_NT') {
                    strFolderDiffOutput = strFolderDiffOutput.replace(oLeftOnlyReplaceRegex, '< <  Only: $1/$2');
                    strFolderDiffOutput = strFolderDiffOutput.replace(oRightOnlyReplaceRegex, ' > > Only: $1/$2');
                }
                else {
                    strFolderDiffOutput = strFolderDiffOutput.replace(oLeftOnlyReplaceRegex, '< <  Only: $1\\$2');
                    strFolderDiffOutput = strFolderDiffOutput.replace(oRightOnlyReplaceRegex, ' > > Only: $1\\$2');
                }
                strFolderDiffOutput = strFolderDiffOutput.replace(/< <  Only: \//gm, '< <  Only: ');
                strFolderDiffOutput = strFolderDiffOutput.replace(/ > > Only: \//gm, ' > > Only: ');
                strFolderDiffOutput = strFolderDiffOutput.replace(/< <  Only: \\/gm, '< <  Only: ');
                strFolderDiffOutput = strFolderDiffOutput.replace(/ > > Only: \\/gm, ' > > Only: ');
                if (strFolderDiffOutput.trim()) {
                    let strArrChoices = strFolderDiffOutput.split('\n');
                    if (strArrChoices && strArrChoices[0].trim().length == 0)
                        strArrChoices.shift();
                    if (strArrChoices && strArrChoices[strArrChoices.length - 1].trim().length == 0)
                        strArrChoices.pop();
                    let oTimerEnd = moment_1.default();
                    let intMilliseconds = oTimerEnd.diff(oTimerStart, 'milliseconds');
                    let intSeconds = Math.round(intMilliseconds / 100) / 10;
                    let intMinutes = Math.round(intMilliseconds / 100 / 60) / 10;
                    let strFileCount = strArrChoices.length.toLocaleString();
                    let strStatusLine = '';
                    if (intSeconds == 1) {
                        strStatusLine = strFileCount + ' files in 1 second.';
                    }
                    else if (intSeconds < 60) {
                        strStatusLine = strFileCount + ' files in ' + intSeconds + ' seconds.';
                    }
                    else if (intSeconds == 60) {
                        strStatusLine = strFileCount + ' files in 1 minute.';
                    }
                    else {
                        strStatusLine = strFileCount + ' files in ' + intMinutes.toLocaleString() + ' minutes.';
                    }
                    spinner.text = strStatusLine;
                    spinner.succeed();
                    this.strArrListOfFileChanges = strArrChoices;
                }
                else {
                    spinner.text = 'No Differences';
                    spinner.succeed();
                }
                fs.unlink(strDiffProgramExcludeFilename, (err) => {
                    if (err)
                        console.log(err.message);
                });
                this.showFolderDifferences();
            });
        }
    }
    async showFolderDifferences() {
        console.log('\n\n');
        let strSelectedLine = await this.askForListChoiceFuzzy('Select a file:', this.strArrListOfFileChanges);
        let strLeftFile = '';
        let strRightFile = '';
        let strSelectedFile = strSelectedLine;
        // if (os.type() === 'Windows_NT') {
        if (/^Added: /.test(strSelectedLine) || /^< <  Only: /.test(strSelectedLine)) {
            strSelectedFile = strSelectedFile.replace(/^Added: /, '');
            strSelectedFile = strSelectedFile.replace(/^< <  Only: /, '');
            strLeftFile = path_1.default.join(this.strLeftPath, strSelectedFile);
            strRightFile = path_1.default.join(this.strRightPath, strSelectedFile);
            let oCliTable = new cli_table3_1.default({ style: { head: [], border: [] } });
            oCliTable.push([{ colSpan: 2, content: chalk.default.cyanBright('Press a Key to Select Action:') }], [chalk.default.greenBright('d'), 'Delete ' + strLeftFile], [chalk.default.greenBright('c'), 'Copy file to ' + strRightFile], [chalk.default.greenBright('f'), 'Back to Files']);
            console.log('\n\n');
            console.log(oCliTable.toString());
            let strLetter = await clikey_1.default('?');
            console.log('>' + strLetter);
            if (strLetter == 'd') {
                this.backupFile(strLeftFile);
                fs_extra_1.default.unlinkSync(strLeftFile);
                this.loadFolderDifferences();
            }
            else if (strLetter == 'c') {
                this.backupFile(strLeftFile);
                fs_extra_1.default.copyFileSync(strLeftFile, strRightFile);
                this.loadFolderDifferences();
            }
            else if (strLetter == 'f') {
                this.showFolderDifferences();
            }
        }
        else if (/^Extra: /.test(strSelectedLine) || /^ > > Only: /.test(strSelectedLine)) {
            strSelectedFile = strSelectedFile.replace(/^Extra: /, '');
            strSelectedFile = strSelectedFile.replace(/^ > > Only: /, '');
            strLeftFile = path_1.default.join(this.strLeftPath, strSelectedFile);
            strRightFile = path_1.default.join(this.strRightPath, strSelectedFile);
            let oCliTable = new cli_table3_1.default({ style: { head: [], border: [] } });
            oCliTable.push([{ colSpan: 2, content: chalk.default.cyanBright('Press a Key to Select Action:') }], [chalk.default.greenBright('d'), 'Delete ' + strRightFile], [chalk.default.greenBright('c'), 'Copy file to ' + strLeftFile], [chalk.default.greenBright('f'), 'Back to Files']);
            console.log('\n\n');
            console.log(oCliTable.toString());
            let strLetter = await clikey_1.default('?');
            console.log('>' + strLetter);
            if (strLetter == 'd') {
                this.backupFile(strRightFile);
                fs_extra_1.default.unlinkSync(strRightFile);
                this.loadFolderDifferences();
            }
            else if (strLetter == 'c') {
                this.backupFile(strRightFile);
                fs_extra_1.default.copyFileSync(strRightFile, strLeftFile);
                this.loadFolderDifferences();
            }
            else if (strLetter == 'f') {
                this.showFolderDifferences();
            }
        }
        else if (/^Older: /.test(strSelectedLine) ||
            /^ *Older: /.test(strSelectedLine) ||
            /^ *Newer: /.test(strSelectedLine) ||
            /^ *Bigger: /.test(strSelectedLine) ||
            /^ *Smaller: /.test(strSelectedLine) ||
            /^Newer: /.test(strSelectedLine) ||
            /^Different: /.test(strSelectedLine) ||
            /^ > >Newer: /.test(strSelectedLine) ||
            /^< < Older: /.test(strSelectedLine)) {
            strSelectedFile = strSelectedFile.replace(/^Older: /, '');
            strSelectedFile = strSelectedFile.replace(/^Newer: /, '');
            strSelectedFile = strSelectedFile.replace(/^ *Newer: /, '');
            strSelectedFile = strSelectedFile.replace(/^ *Older: /, '');
            strSelectedFile = strSelectedFile.replace(/^ *Bigger: /, '');
            strSelectedFile = strSelectedFile.replace(/^ *Smaller: /, '');
            strSelectedFile = strSelectedFile.replace(/^Different: /, '');
            strSelectedFile = strSelectedFile.replace(/^ > >Newer: /, '');
            strSelectedFile = strSelectedFile.replace(/^< < Older: /, '');
            strLeftFile = path_1.default.join(this.strLeftPath, strSelectedFile);
            strRightFile = path_1.default.join(this.strRightPath, strSelectedFile);
            this.showColorDiff(strLeftFile, strRightFile, 1, this.strLeftPath, this.strRightPath);
        }
    }
    formatBytes(a, b) {
        if (0 == a)
            return '0 Bytes';
        var c = 1024, d = b || 2, e = ['Bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'], f = Math.floor(Math.log(a) / Math.log(c));
        return parseFloat((a / Math.pow(c, f)).toFixed(d)) + ' ' + e[f];
    }
    async showColorDiff(strLeftFile, strRightFile, Hunk, FolderOne, FolderTwo) {
        if (Hunk < 1)
            Hunk = 1;
        //
        //git diff --word-diff=color ../diff-patch-tests/1 ../diff-patch-tests/2
        let strDiffProgram = 'diff';
        // let strArrDiffSwitches: string[] = [ '--color=always', '--unified' ]
        let strArrDiffSwitches = ['--unified'];
        // let strDiffProgram: string = 'git'
        // let strArrDiffSwitches: string[] = [ 'diff', '--word-diff=color', '--unified' ]
        strArrDiffSwitches.push(strLeftFile);
        strArrDiffSwitches.push(strRightFile);
        const proc = cross_spawn_1.default(strDiffProgram, strArrDiffSwitches);
        // const proc = spawn('git', strArrDiffSwitches)
        let output = '';
        proc.stdout.on('data', (chunk) => {
            output += chunk.toString();
        });
        proc.on('exit', async () => {
            let strPatchHeaderLine = '';
            console.log(chalk.default.grey.dim.italic('Running: ' + strDiffProgram + ' ' + strArrDiffSwitches.join(' ')));
            // terminal escape codes: \u001b[36m
            // ing.pdf"\r\n\u001b[36m@@ -242,10 +242,10 @@\u001b[0m\n \t\t\t\t\t\t target="_b
            // let strHunks = output.split(/\n\u001b\[36m\@\@ -\d+\,\d+ \+\d+\,\d+ \@\@/)
            // let strHunks = output.split(/\u001b\[36m\@\@ -\d*\,\d* \+\d*\,\d* \@\@/)
            let strHunks = new Array();
            let strHunkForDisplay = '';
            if (output) {
                strHunks = output.split(/^@\@ -/gm);
            }
            if (strHunks.length > 0) {
                strPatchHeaderLine = strHunks[0];
                strHunks.shift();
                if (Hunk > strHunks.length) {
                    Hunk = strHunks.length;
                }
            }
            if (output) {
                strHunkForDisplay = strHunks[Hunk - 1];
            }
            let intLeftStartLine = 0;
            let intRightStartLine = 0;
            let intLeftNumberOfLines = 0;
            let intRightNumberOfLines = 0;
            const regexLineNumberMatches = /^([0-9]+),([0-9]+) \+([0-9]+),([0-9]+) @@/gm.exec(strHunkForDisplay);
            if (strHunkForDisplay.trim()) {
                if (regexLineNumberMatches[1] &&
                    regexLineNumberMatches[2] &&
                    regexLineNumberMatches[3] &&
                    regexLineNumberMatches[4]) {
                    if (/^[0-9]+$/.test(regexLineNumberMatches[1])) {
                        intLeftStartLine = Number(regexLineNumberMatches[1]);
                    }
                    if (/^[0-9]+$/.test(regexLineNumberMatches[2])) {
                        intLeftNumberOfLines = Number(regexLineNumberMatches[2]);
                    }
                    if (/^[0-9]+$/.test(regexLineNumberMatches[3])) {
                        intRightStartLine = Number(regexLineNumberMatches[3]);
                    }
                    if (/^[0-9]+$/.test(regexLineNumberMatches[4])) {
                        intRightNumberOfLines = Number(regexLineNumberMatches[4]);
                    }
                }
            }
            // start red
            // [31m
            // end red
            // [0m
            // start yellow
            // [32m
            // end yellow
            // [0m
            strHunkForDisplay = strHunkForDisplay.replace(/^\d*\,\d* \+\d*\,\d* \@\@/, '');
            let strHunkForDisplayLines = strHunkForDisplay.split('\n');
            let strHunkForDisplayLinesA = new Array();
            let strHunkForDisplayLinesB = new Array();
            let booYellow = false;
            for (let strLine of strHunkForDisplayLines) {
                if (strLine.startsWith('+')) {
                    // yellow
                    strLine = strLine.substr(1);
                    strHunkForDisplayLinesB.push(strLine);
                    booYellow = true;
                }
                else if (strLine.startsWith('-')) {
                    // red
                    strLine = strLine.substr(1);
                    if (booYellow) {
                        if (strHunkForDisplayLinesA.length > strHunkForDisplayLinesB.length) {
                            do {
                                strHunkForDisplayLinesB.push('');
                            } while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length);
                        }
                        else if (strHunkForDisplayLinesA.length < strHunkForDisplayLinesB.length) {
                            do {
                                strHunkForDisplayLinesA.push('');
                            } while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length);
                        }
                    }
                    strHunkForDisplayLinesA.push(strLine);
                    booYellow = false;
                }
                else {
                    if (booYellow) {
                        if (strHunkForDisplayLinesA.length > strHunkForDisplayLinesB.length) {
                            do {
                                strHunkForDisplayLinesB.push('');
                            } while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length);
                        }
                        else if (strHunkForDisplayLinesA.length < strHunkForDisplayLinesB.length) {
                            do {
                                strHunkForDisplayLinesA.push('');
                            } while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length);
                        }
                    }
                    strHunkForDisplayLinesA.push(strLine);
                    strHunkForDisplayLinesB.push(strLine);
                    booYellow = false;
                }
            }
            let strLeft = '';
            let strRight = '';
            let strDiffLR = '';
            let strDiffRL = '';
            let diffL;
            let diffR;
            for (let i = 0; i < strHunkForDisplayLinesA.length - 1; i++) {
                strLeft = strHunkForDisplayLinesA[i];
                strRight = strHunkForDisplayLinesB[i];
                if (strLeft != strRight) {
                    diffL = diff_1.default.diffWordsWithSpace(strLeft, strRight);
                    strDiffLR = '';
                    strDiffRL = '';
                    diffL.forEach(function (part) {
                        if (part.added) {
                        }
                        else if (part.removed) {
                            strDiffLR = strDiffLR + chalk.default.black.bgGreen(part.value);
                        }
                        else {
                            strDiffLR = strDiffLR + part.value;
                        }
                    });
                    strHunkForDisplayLinesA[i] = strDiffLR;
                    diffR = diff_1.default.diffWordsWithSpace(strRight, strLeft);
                    diffR.forEach(function (part) {
                        if (part.added) {
                        }
                        else if (part.removed) {
                            strDiffRL = strDiffRL + chalk.default.black.bgRed(part.value);
                        }
                        else {
                            strDiffRL = strDiffRL + part.value;
                        }
                    });
                    strHunkForDisplayLinesB[i] = strDiffRL;
                }
            }
            // for (let strLine of strHunkForDisplayLines) {
            // 	if (strLine.startsWith('+')) {
            // 		// yellow
            // 		strHunkForDisplayLinesB.push('[32m' + strLine + '[0m')
            // 		booYellow = true
            // 	} else if (strLine.startsWith('[32m')) {
            // 		// yellow
            // 		strHunkForDisplayLinesB.push(strLine + '[0m')
            // 		booYellow = true
            // 	} else if (strLine.startsWith('-')) {
            // 		// red
            // 		if (booYellow) {
            // 			if (strHunkForDisplayLinesA.length > strHunkForDisplayLinesB.length) {
            // 				do {
            // 					strHunkForDisplayLinesB.push('')
            // 				} while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length)
            // 			} else if (strHunkForDisplayLinesA.length < strHunkForDisplayLinesB.length) {
            // 				do {
            // 					strHunkForDisplayLinesA.push('')
            // 				} while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length)
            // 			}
            // 		}
            // 		strHunkForDisplayLinesA.push('[31m' + strLine + '[0m')
            // 		booYellow = false
            // 	} else if (strLine.startsWith('[31m')) {
            // 		// red
            // 		if (booYellow) {
            // 			if (strHunkForDisplayLinesA.length > strHunkForDisplayLinesB.length) {
            // 				do {
            // 					strHunkForDisplayLinesB.push('')
            // 				} while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length)
            // 			} else if (strHunkForDisplayLinesA.length < strHunkForDisplayLinesB.length) {
            // 				do {
            // 					strHunkForDisplayLinesA.push('')
            // 				} while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length)
            // 			}
            // 		}
            // 		strHunkForDisplayLinesA.push(strLine + '[0m')
            // 		booYellow = false
            // 	} else {
            // 		if (booYellow) {
            // 			if (strHunkForDisplayLinesA.length > strHunkForDisplayLinesB.length) {
            // 				do {
            // 					strHunkForDisplayLinesB.push('')
            // 				} while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length)
            // 			} else if (strHunkForDisplayLinesA.length < strHunkForDisplayLinesB.length) {
            // 				do {
            // 					strHunkForDisplayLinesA.push('')
            // 				} while (strHunkForDisplayLinesA.length != strHunkForDisplayLinesB.length)
            // 			}
            // 		}
            // 		strHunkForDisplayLinesA.push(strLine)
            // 		strHunkForDisplayLinesB.push(strLine)
            // 		booYellow = false
            // 	}
            // }
            let intMaxLength;
            if (intLeftNumberOfLines > intRightNumberOfLines) {
                intMaxLength = intLeftNumberOfLines;
            }
            else {
                intMaxLength = intRightNumberOfLines;
            }
            let intColWidth = Math.floor(this.terminalSize.columns / 2 - 5);
            let strLeftFolderFilePath = path_1.default.dirname(strLeftFile);
            let strRightFolderFilePath = path_1.default.dirname(strRightFile);
            let strLeftFilename = path_1.default.basename(strLeftFile);
            let strRightFilename = path_1.default.basename(strRightFile);
            let strLeftModeTime;
            let strRightModeTime;
            let strLeftSize;
            let strRightSize;
            if (fs_extra_1.default.existsSync(strLeftFile)) {
                strLeftModeTime =
                    fs_extra_1.default.statSync(strLeftFile).mtime.toLocaleDateString() +
                        ' ' +
                        fs_extra_1.default.statSync(strLeftFile).mtime.toLocaleTimeString();
                strLeftSize = this.formatBytes(fs_extra_1.default.statSync(strLeftFile).size, 3);
            }
            if (fs_extra_1.default.existsSync(strRightFile)) {
                strRightModeTime =
                    fs_extra_1.default.statSync(strRightFile).mtime.toLocaleDateString() +
                        ' ' +
                        fs_extra_1.default.statSync(strRightFile).mtime.toLocaleTimeString();
                strRightSize = this.formatBytes(fs_extra_1.default.statSync(strRightFile).size, 3);
            }
            let oCliDiffTable = new cli_table3_1.default({
                head: [
                    '',
                    chalk.default.cyanBright('Folder: ' +
                        strLeftFolderFilePath +
                        '\n' +
                        'File: ' +
                        strLeftFilename +
                        '\n' +
                        'line ' +
                        intLeftStartLine.toLocaleString() +
                        ' to ' +
                        (intLeftStartLine + (intLeftNumberOfLines - 2)).toLocaleString() +
                        '\nModified: ' +
                        strLeftModeTime +
                        '\nSize: ' +
                        strLeftSize),
                    chalk.default.cyanBright('Folder: ' +
                        strRightFolderFilePath +
                        '\n' +
                        'File: ' +
                        strRightFilename +
                        '\n' +
                        'line ' +
                        intRightStartLine.toLocaleString() +
                        ' to ' +
                        (intRightStartLine + (intRightNumberOfLines - 2)).toLocaleString() +
                        '\nModified: ' +
                        strRightModeTime +
                        '\nSize: ' +
                        strRightSize),
                ],
                colWidths: [5, intColWidth, intColWidth],
                wordWrap: true,
            });
            let strA;
            let strB;
            let intLine = intLeftStartLine;
            for (let i = 1; i < intMaxLength; i++) {
                strA = '';
                strB = '';
                if (strHunkForDisplayLinesA[i]) {
                    strA = strHunkForDisplayLinesA[i];
                    strA = strA.replace(/\t/gm, '  ');
                    strA = strA.replace(/\r/gm, '');
                }
                if (strHunkForDisplayLinesB[i]) {
                    strB = strHunkForDisplayLinesB[i];
                    strB = strB.replace(/\t/gm, '  ');
                    strB = strB.replace(/\r/gm, '');
                }
                oCliDiffTable.push([intLine, strA, strB]);
                intLine++;
            }
            if (strHunks && strHunks.length > 0)
                console.log(oCliDiffTable.toString());
            // console.log(strHunkForDisplay)
            let oCliTable = new cli_table3_1.default({ style: { head: [], border: [] } });
            oCliTable.push([{ colSpan: 4, content: chalk.default.cyanBright('Press a Key to Select Action:') }]);
            if (strHunks.length == 1) {
                // This is the only one.
            }
            else if (strHunks.length > 0 && Hunk == 1) {
                // This is not the only one but it is the first one.
                oCliTable.push([chalk.default.greenBright('n'), 'Next Difference', '', '']);
            }
            else if (strHunks.length > 0 && Hunk == strHunks.length) {
                // Last on in the list
                oCliTable.push([chalk.default.greenBright('p'), 'Previous Difference', '', '']);
            }
            else if (strHunks.length == 0) {
                // if (os.type() === 'Windows_NT') {
                // Robocopy checks the date of tiles even it they are not different
                // This showed up on the list but there are no differences.  So, it must be because of the timestamp.
                let dtm1MTime = fs.statSync(strLeftFile).mtime;
                let dtm2MTime = fs.statSync(strRightFile).mtime;
                let nTimeDiffInSeconds = Math.abs((dtm1MTime.getTime() - dtm2MTime.getTime()) / 1000);
                if (nTimeDiffInSeconds > 1)
                    oCliTable.push([
                        chalk.default.greenBright('t'),
                        {
                            colSpan: 3,
                            content: 'Make Timestamps Equal\nLeft=' +
                                dtm1MTime.toLocaleDateString() +
                                ' ' +
                                dtm1MTime.toLocaleTimeString() +
                                '\nRight=' +
                                dtm2MTime.toLocaleDateString() +
                                ' ' +
                                dtm2MTime.toLocaleTimeString() +
                                '\nSet both to Right',
                        },
                    ]);
                // }
            }
            else {
                oCliTable.push([
                    chalk.default.greenBright('n'),
                    'Next Difference',
                    chalk.default.greenBright('p'),
                    'Previous Difference',
                ]);
            }
            if (strHunks.length != 0) {
                // There were hunks returned
                oCliTable.push([
                    chalk.default.greenBright('a'),
                    'Apply Left to Right',
                    chalk.default.greenBright('b'),
                    'Apply Right to Left',
                ], [
                    chalk.default.greenBright('A'),
                    'Copy File from Left to Right',
                    chalk.default.greenBright('B'),
                    'Copy File from Right to Left',
                ], [chalk.default.greenBright('e'), 'Edit Left', chalk.default.greenBright('E'), 'Edit Right']);
            }
            oCliTable.push([
                chalk.default.greenBright('f'),
                'Back to Files',
                chalk.default.greenBright('r'),
                'Restart',
            ]);
            // console.log('\u2500'.repeat(40))
            if (strHunks.length) {
                console.log('Difference ' + Hunk.toString() + ' of ' + strHunks.length.toString() + '.');
            }
            else {
                console.log('No Difference.');
            }
            // console.log('\u2500'.repeat(40))
            console.log(oCliTable.toString());
            let strLetter = await clikey_1.default('?');
            console.log('>' + strLetter);
            if (strLetter == 'r') {
                this.selectCompareFolders();
            }
            else if (strLetter == 'e') {
                this.backupFile(strLeftFile);
                if (intLeftStartLine) {
                    this, this.editFileSync(strLeftFile, intLeftStartLine);
                }
                else {
                    this, this.editFileSync(strLeftFile);
                }
                this.showColorDiff(strLeftFile, strRightFile, Hunk + 1, FolderOne, FolderTwo);
            }
            else if (strLetter == 'E') {
                this.backupFile(strRightFile);
                if (intRightStartLine) {
                    this, this.editFileSync(strRightFile, intRightStartLine);
                }
                else {
                    this, this.editFileSync(strRightFile);
                }
                this.showColorDiff(strLeftFile, strRightFile, Hunk + 1, FolderOne, FolderTwo);
            }
            else if (strLetter == 'n') {
                this.showColorDiff(strLeftFile, strRightFile, Hunk + 1, FolderOne, FolderTwo);
            }
            else if (strLetter == 'p') {
                this.showColorDiff(strLeftFile, strRightFile, Hunk - 1, FolderOne, FolderTwo);
            }
            else if (strLetter == 'f') {
                console.log('\n\n');
                console.log('Rescan folders for changes: y/N');
                let strLetter = await clikey_1.default('?');
                console.log('>' + strLetter);
                if (strLetter.toLowerCase() == 'y') {
                    this.loadFolderDifferences();
                }
                else {
                    this.showFolderDifferences();
                }
            }
            else if (strLetter == 't') {
                let dtrATime = fs.statSync(strRightFile).atime;
                let dtrMTime = fs.statSync(strRightFile).mtime;
                console.log('Setting last access time on ' +
                    strLeftFile +
                    ' to ' +
                    dtrATime.toLocaleDateString() +
                    ' ' +
                    dtrATime.toLocaleTimeString() +
                    ' and the modified time to ' +
                    dtrMTime.toLocaleDateString() +
                    ' ' +
                    dtrMTime.toLocaleTimeString());
                fs.utimesSync(strLeftFile, fs.statSync(strRightFile).atime, fs.statSync(strRightFile).mtime);
                console.log('\n\n');
                console.log('Rescan folders for changes: y/N');
                let strLetter = await clikey_1.default('?');
                console.log('>' + strLetter);
                if (strLetter.toLowerCase() == 'y') {
                    this.loadFolderDifferences();
                }
                else {
                    this.showFolderDifferences();
                }
            }
            else if (strLetter == 'B') {
                this.backupFile(strLeftFile);
                fs_extra_1.default.copyFileSync(strRightFile, strLeftFile);
                this.showColorDiff(strLeftFile, strRightFile, Hunk - 2, FolderOne, FolderTwo);
            }
            else if (strLetter == 'b') {
                this.backupFile(strLeftFile);
                let strPatchString = strPatchHeaderLine + '@@ -' + strHunks[Hunk - 1];
                strPatchString = strPatchString.replace(/\x1b\[[0-9;]*m/g, '');
                temp.track();
                temp.open('interactive-diff-patch---', (err, info) => {
                    if (!err) {
                        fs.writeFile(info.fd, strPatchString, (err) => {
                            if (err)
                                throw err;
                            fs.close(info.fd, (err) => {
                                if (err)
                                    throw err;
                                let strCommand = 'patch';
                                let strArrPatchSwitches = [
                                    '--strip=0',
                                    '--directory=/',
                                    '--input=' + info.path,
                                    strLeftFile,
                                ];
                                console.log(chalk.default.grey.dim.italic('Running: patch ' + strArrPatchSwitches.join(' ')));
                                const proc = cross_spawn_1.default(strCommand, strArrPatchSwitches);
                                proc.on('exit', () => {
                                    temp.cleanupSync();
                                    fs.utimesSync(strLeftFile, fs.statSync(strRightFile).atime, fs.statSync(strRightFile).mtime);
                                    setTimeout(() => {
                                        this.showColorDiff(strLeftFile, strRightFile, Hunk - 2, FolderOne, FolderTwo);
                                    }, lodash_1.random(30, 500));
                                });
                            });
                        });
                    }
                });
            }
            else if (strLetter == 'A') {
                this.backupFile(strRightFile);
                fs_extra_1.default.copyFileSync(strLeftFile, strRightFile);
                this.showColorDiff(strLeftFile, strRightFile, Hunk - 2, FolderOne, FolderTwo);
            }
            else if (strLetter == 'a') {
                this.backupFile(strRightFile);
                let strPatchString = strPatchHeaderLine + '@@ -' + strHunks[Hunk - 1];
                strPatchString = strPatchString.replace(/\x1b\[[0-9;]*m/g, '');
                temp.track();
                temp.open('interactive-diff-patch---', (err, info) => {
                    if (!err) {
                        fs.writeFile(info.fd, strPatchString, (err) => {
                            if (err)
                                throw err;
                            fs.close(info.fd, (err) => {
                                if (err)
                                    throw err;
                                let strArrPatchSwitches = [
                                    '--strip=0',
                                    '--directory=/',
                                    '--reverse',
                                    '--input=' + info.path,
                                    strRightFile,
                                ];
                                console.log(chalk.default.grey.dim.italic('Running: patch ' + strArrPatchSwitches.join(' ')));
                                const proc = cross_spawn_1.default('patch', strArrPatchSwitches);
                                // let output = ''
                                // proc.stdout.on('data', (chunk) => {
                                // 	output += chunk.toString()
                                // })
                                proc.on('exit', async () => {
                                    temp.cleanupSync();
                                    fs.utimesSync(strRightFile, fs.statSync(strLeftFile).atime, fs.statSync(strLeftFile).mtime);
                                    setTimeout(() => {
                                        this.showColorDiff(strLeftFile, strRightFile, Hunk - 2, FolderOne, FolderTwo);
                                    }, lodash_1.random(30, 500));
                                    // console.log(output)
                                });
                            });
                        });
                    }
                });
                // console.log(strPatchString)
            }
        });
    }
    async askForString(Question, Default) {
        return inquirer
            .prompt([
            {
                type: 'input',
                message: Question,
                name: 'reply',
                default: function () {
                    return Default;
                },
                filter: (input) => {
                    return input;
                },
            },
        ])
            .then((input) => {
            return input.reply;
        });
    }
    async askForListChoice(Message, Choices) {
        console.log(Message);
        return inquirer
            .prompt([
            {
                type: 'list',
                name: 'val',
                suggestOnly: false,
                message: Message,
                choices: Choices,
                pageSize: 20,
            },
        ])
            .then(async (answers) => {
            return answers.val;
        });
    }
    async askForListChoiceFuzzy(Message, Choices) {
        console.log(Message);
        inquirer.registerPrompt('autocomplete', inquirer_autocomplete_prompt_1.default);
        return inquirer
            .prompt([
            {
                type: 'autocomplete',
                name: 'val',
                suggestOnly: false,
                message: Message,
                source: (answers, strUserInput) => {
                    strUserInput = strUserInput || '';
                    return new Promise((resolve, reject) => {
                        setTimeout(() => {
                            let fuzzyResult = fuzzy_1.filter(strUserInput, Choices);
                            resolve(fuzzyResult.map((el) => {
                                return el.original;
                            }));
                        }, lodash_1.random(30, 500));
                    });
                },
                pageSize: 20,
            },
        ])
            .then(async (answers) => {
            return answers.val;
        });
    }
    async askForListChoiceFuzzyForFiles(Message, Choices) {
        inquirer.registerPrompt('autocomplete', inquirer_autocomplete_prompt_1.default);
        return inquirer
            .prompt([
            {
                type: 'autocomplete',
                name: 'val',
                suggestOnly: false,
                message: Message,
                pageSize: 20,
                onKeypress: '',
                source: (answers, strUserInput) => {
                    strUserInput = strUserInput || '';
                    return new Promise((resolve, reject) => {
                        setTimeout(() => {
                            let fuzzyResult = fuzzy_1.filter(strUserInput, Choices);
                            resolve(fuzzyResult.map((el) => {
                                return el.original;
                            }));
                        }, lodash_1.random(30, 500));
                    });
                },
            },
        ])
            .then(async (answers) => {
            return answers.val;
        });
    }
    async askForFolder(Message, BaseFolderPath) {
        inquirer.registerPrompt('directory', inquirer_directory_1.default);
        let strPath = BaseFolderPath;
        return inquirer
            .prompt([
            {
                type: 'directory',
                name: 'path',
                message: Message,
                basePath: strPath,
                options: {
                    displayFiles: false,
                    displayHidden: true,
                },
            },
        ])
            .then(function (answer) {
            return answer.path;
        });
    }
    getWebPage(url) {
        return new Promise((resolve, reject) => {
            https_1.default
                .get(url, (resp) => {
                let data = '';
                resp.on('data', (chunk) => {
                    data += chunk;
                });
                resp.on('end', () => {
                    resolve(data);
                });
            })
                .on('error', (err) => {
                console.log('Error: ' + err.message);
            });
        });
    }
    editFileSync(Filename, LineNumber) {
        let strCommand = 'nano';
        if (os.type() === 'Windows_NT') {
            try {
                if (fs.existsSync(this.strNanoGitWinPath)) {
                    strCommand = '"' + this.strNanoGitWinPath + '"';
                }
                else {
                    let strAppDataFolder = appdata_path_1.default('interactive-diff-patch');
                    let strDownloadFilePath = strAppDataFolder + '/nano.exe';
                    strCommand = '"' + strDownloadFilePath + '"';
                    if (!fs.existsSync(strDownloadFilePath)) {
                        strCommand = 'notepad';
                    }
                }
            }
            catch (error) {
                strCommand = 'notepad';
            }
        }
        if (LineNumber) {
            if (strCommand != 'notepad') {
                strCommand = strCommand + ' -l ';
            }
            child_process_1.execSync(strCommand + ' +' + LineNumber.toString() + ' ' + Filename, { stdio: [0, 1, 2] });
        }
        else {
            child_process_1.execSync(strCommand + ' ' + Filename, { stdio: [0, 1, 2] });
        }
    }
    async compareFoldersSync(Folder) { }
}
new DiffPatch();
// making aur files
// npm2PKGBUILD interactive-diff-patch > PKGBUILD
// makepkg --printsrcinfo > .SRCINFO
// add find "${pkgdir}"/usr -type d -exec chmod 755 {} +
// to last line of package section in PKGBUILD
// dpkg-buildpackage
// mk-build-deps -i debian/control
// sudo dpkg-sig --sign builder mypackage.deb
//# sourceMappingURL=index.js.map